@@ -2,9 +2,7 @@ package Constant::Export::Lazy;
 BEGIN {
   $Constant::Export::Lazy::AUTHORITY = 'cpan:AVAR';
 }
-{
-  $Constant::Export::Lazy::VERSION = '0.03';
-}
+$Constant::Export::Lazy::VERSION = '0.04';
 use strict;
 use warnings;
 
@@ -17,21 +15,21 @@ sub import {
         exists $args{options}
         ? exists $args{options}->{wrap_existing_import}
           ? $args{options}->{wrap_existing_import}
-          : 0
-        : 0
+          : undef
+        : undef
     );
     my $existing_import;
     my $caller_import_name = $caller . '::import';
 
     # Sanity check whether we do or don't have an existing 'import'
     # sub with the wrap_existing_import option:
-    my $has_import_already = do { no strict 'refs'; *{$caller_import_name}{CODE} } ? 1 : 0;
+    my $has_import_already = do { no strict 'refs'; no warnings 'once'; *{$caller_import_name}{CODE} } ? 1 : 0;
     {
         if ($wrap_existing_import) {
             die "PANIC: We need an existing 'import' with the wrap_existing_import option" unless $has_import_already;
             $existing_import = \&{$caller_import_name};
         } else {
-            die "PANIC: We're trying to clobber some existing 'import' subroutine without having the 'wrap_existing_import' option" if $has_import_already;
+            die "PANIC: We're trying to clobber an existing 'import' subroutine without having the 'wrap_existing_import' option" if $has_import_already;
         }
     }
 
@@ -42,6 +40,16 @@ sub import {
     my $normalized_args = _normalize_arguments(%args);
     my $constants = $normalized_args->{constants};
 
+    # This is a callback that can be used to munge the import list, to
+    # e.g. provide a facility to provide import tags.
+    my $buildargs = (
+        exists $args{options}
+        ? exists $args{options}->{buildargs}
+          ? $args{options}->{buildargs}
+          : undef
+        : undef
+    );
+
     no strict 'refs';
     no warnings 'redefine'; # In case of $wrap_existing_import
     *{$caller_import_name} = sub {
@@ -85,8 +93,25 @@ sub import {
                 no strict 'refs';
                 \%{"${caller}::"};
             },
+
+            # If we're not wrapping an existing import subroutine we
+            # don't need to bend over backwards to support constants
+            # generated by e.g. constant.pm, we know we've made all
+            # the constants in the package to our liking.
+            wrap_existing_import => $wrap_existing_import,
         );
 
+        # We've been provided with a callback to be used to munge
+        # whatever we actually got provided with in @gimme to a list
+        # of constants, or if $wrap_existing_import is enabled any
+        # leftover non-$gimme names it's going to handle.
+        if ($buildargs) {
+            my @overriden_gimme = $buildargs->(\@gimme, $constants);
+            die "PANIC: We only support subs that return zero or one values with buildargs, yours returns " . @overriden_gimme . " values"
+                if @overriden_gimme > 1;
+            @gimme = @{$overriden_gimme[0]} if @overriden_gimme;
+        }
+
         # Just doing ->call() like you would when you're using the API
         # will fleshen the constant, do this for all the constants
         # we've been requested to export.
@@ -153,9 +178,7 @@ package Constant::Export::Lazy::Ctx;
 BEGIN {
   $Constant::Export::Lazy::Ctx::AUTHORITY = 'cpan:AVAR';
 }
-{
-  $Constant::Export::Lazy::Ctx::VERSION = '0.03';
-}
+$Constant::Export::Lazy::Ctx::VERSION = '0.04';
 use strict;
 use warnings;
 
@@ -172,27 +195,68 @@ sub call {
     my ($ctx, $gimme) = @_;
 
     # Unpack our options
-    my $symtab       = $ctx->{symtab};
-    my $pkg_importer = $ctx->{pkg_importer};
-    my $pkg_stash    = $ctx->{pkg_stash};
-    my $constants    = $ctx->{constants};
+    my $symtab               = $ctx->{symtab};
+    my $pkg_importer         = $ctx->{pkg_importer};
+    my $pkg_stash            = $ctx->{pkg_stash};
+    my $constants            = $ctx->{constants};
+    my $wrap_existing_import = $ctx->{wrap_existing_import};
 
-    my $glob_name = "${pkg_stash}::${gimme}";
-    my $alias_as  = "${pkg_importer}::${gimme}";
+    # Unless we're wrapping an existing import ->call($gimme) should
+    # always be called with a $gimme that we know about.
+    unless (exists $constants->{$gimme}) {
+        die "PANIC: You're trying to get the value of an unknown constant, and wrap_existing_import isn't set" unless $wrap_existing_import;
+    }
+
+    my ($private_name, $glob_name, $alias_as);
+    my $make_private_glob_and_alias_name = sub {
+        # Checking "exists $constants->{$gimme}" here to avoid
+        # autovivification would be redundant since we'll die before
+        # we get here unless $wrap_existing_import is true above. If
+        # $wrap_existing_import is true we'll have called the import()
+        # we're wrapping if it's a constant we don't know about, or
+        # it's being called from ->call(), in which case
+        # $CALL_SHOULD_NOT_ALIAS is true and we won't be calling this
+        # sub unless $constants->{$gimme} exists. Since this won't
+        $private_name = exists $constants->{$gimme}->{options}->{private_name_munger}
+                             ? $constants->{$gimme}->{options}->{private_name_munger}->($gimme)
+                             : $gimme;
+        $private_name = defined $private_name ? $private_name : $gimme;
+        $glob_name = "${pkg_stash}::${private_name}";
+        $alias_as  = "${pkg_importer}::${gimme}";
+        return;
+    };
 
     my $value;
-    if (exists $symtab->{$gimme}) {
+    if ($wrap_existing_import and not exists $constants->{$gimme}) {
+        # This is in case $ctx->call() is used on a constant defined
+        # by constant.pm. See the giant comment about constant.pm
+        # below.
         my $symtab_value = $symtab->{$gimme};
-        if (ref $symtab_value eq 'SCALAR') {
-            # This is in case $ctx->call() is used on a constant
-            # defined by constant.pm. See the giant comment about
-            # constant.pm below.
+        my $ref = ref $symtab_value;
+        if (not defined $symtab_value) {
+            die "PANIC: We're trying to fallback to a constant we don't know about under wrap_existing_import, but $gimme has no symbol table entry";
+        } elsif ($ref eq 'SCALAR' or $ref eq 'REF') {
             $value = $$symtab_value;
+        } elsif (ref $symtab_value eq 'ARRAY') {
+            $value = $symtab_value;
         } else {
-            # TODO: Is there some better way to check if this is a
-            # code value than just "ne 'SCALAR'"?
-            $value = &$symtab_value();
+            my @value = &$symtab_value();
+            die "PANIC: We only support subs that return one value with wrap_existing_import, $gimme returns " . @value . " values"
+                if @value > 1;
+            $value = $value[0];
         }
+    } elsif (do {
+        # If we got this far we know we're going to want to call
+        # $make_private_glob_and_alias_name->(). It'll also be used by
+        # branches further down.
+        $make_private_glob_and_alias_name->();
+
+        exists $symtab->{$private_name};
+    }) {
+        # This is for constants that *we've* previously defined, we'll
+        # always use our own $private_name.
+        my $symtab_value = $symtab->{$private_name};
+        $value = &$symtab_value();
     } else {
         # We only want to alias constants into the importer's package
         # if the constant is on the import list, not if it's just
@@ -250,18 +314,17 @@ sub call {
             #  * perl-5.005_02-225-g779c5bc - first core support
             #    for these kinds of constants in the optree.
             #
-            # * perl-5.8.0-6623-ge040ff7 - first use in
-            #   constant.pm.
+            # * perl-5.9.2-1966-ge040ff7 - first use in constant.pm.
             #
-            # * perl-5.8.0-6638-ge1234d8 - first attempts to
+            # * perl-5.9.2-1981-ge1234d8 - first attempts to
             #   invalidate the method cache with
             #   Internals::inc_sub_generation()
             #
-            # * perl-5.8.0-10189-ge1a479c -
+            # * perl-5.9.4-1684-ge1a479c -
             #   Internals::inc_sub_generation() in constant.pm
             #   replaced with mro::method_changed_in($pkg)
             #
-            #  * perl-5.8.0-10219-g41892db - Now unused
+            #  * perl-5.9.4-1714-g41892db - Now unused
             #    Internals::inc_sub_generation() removed from the
             #    core.
             #
@@ -302,7 +365,8 @@ sub call {
 
     unless ($CALL_SHOULD_NOT_ALIAS) {
         no strict 'refs';
-        # Alias e.g. user::CONSTANT to YourExporter::CONSTANT
+        # Alias e.g. user::CONSTANT to YourExporter::CONSTANT (well,
+        # actually YourExporter::$private_name)
         *$alias_as = *$glob_name;
     }
 
@@ -333,9 +397,9 @@ Constant::Export::Lazy - Utility to write lazy exporters of constant subroutines
 
 =head1 SYNOPSIS
 
-This increasingly verbose example of your C<My::Constants> package
-that you write using C<Constant::Export::Lazy> demonstrates all our
-features (from F<t/lib/My/Constants.pm> in the source distro):
+These increasing verbose example of a C<My::Constants> package that
+you can write using C<Constant::Export::Lazy> demonstrates most of our
+major features (from F<t/lib/My/Constants.pm> in the source distro):
 
     package My::Constants;
     use strict;
@@ -374,10 +438,8 @@ features (from F<t/lib/My/Constants.pm> in the source distro):
             # thus We'll return [3..4] here.
             #
             # Unlike the constant.pm that ships with perl itself we don't
-            # support returning lists (there's no such things as constant
-            # list subroutines, constant.pm fakes it with a non-inlined
-            # sub). So if you want to return lists you have to return a
-            # reference to one.
+            # support returning lists. So if you want to return lists you
+            # have to return a reference to one.
             LIST => sub { wantarray ? (1..2) : [3..4] },
             # We can also supply a HashRef with "call" with the sub, and
             # "options" with options that clobber the global
@@ -486,6 +548,106 @@ And running it gives:
     ok 8
     1..8
 
+By default we only support importing constants explicitly by their own
+names and not something like L<Exporter>'s C<@EXPORT>, C<@EXPORT_OK>
+or C<%EXPORT_TAGS>, but you can trivially add support for that (or any
+other custom import munging) using the L</buildargs> callback. This
+example is from F<t/lib/My/Constants/Tags.pm> in the source distro:
+
+    package My::Constants::Tags;
+    use v5.10;
+    use strict;
+    use warnings;
+    use Constant::Export::Lazy (
+        constants => {
+            KG_TO_MG => sub { 10**6 },
+            SQRT_2 => {
+                call    => sub { sqrt(2) },
+                options => {
+                    stash => {
+                        export_tags => [ qw/:math/ ],
+                    },
+                },
+            },
+            PI => {
+                call    => sub { atan2(1,1) * 4 },
+                options => {
+                    stash => {
+                        export_tags => [ qw/:math/ ],
+                    },
+                },
+            },
+            map(
+                {
+                    my $t = $_;
+                    +(
+                        $_ => {
+                            call => sub { $t },
+                            options => {
+                                stash => {
+                                    export_tags => [ qw/:alphabet/ ],
+                                },
+                            }
+                        }
+                    )
+                }
+                "A".."Z"
+            ),
+        },
+        options => {
+            buildargs => sub {
+                my ($import_args, $constants) = @_;
+
+                state $export_tags = do {
+                    my %export_tags;
+                    for my $constant (keys %$constants) {
+                        my @export_tags = @{$constants->{$constant}->{options}->{stash}->{export_tags} || []};
+                        push @{$export_tags{$_}} => $constant for @export_tags;
+                    }
+                    \%export_tags;
+                };
+
+                my @gimme = map {
+                    /^:/ ? @{$export_tags->{$_}} : $_
+                } @$import_args;
+
+                return \@gimme;
+            },
+        },
+    );
+
+    1;
+
+And this is an example of using it in some user code (from
+F<t/synopsis_tags.t> in the source distro):
+
+    package My::More::User::Code;
+    use strict;
+    use warnings;
+    use Test::More qw(no_plan);
+    use lib 't/lib';
+    use My::Constants::Tags qw(
+        KG_TO_MG
+        :math
+        :alphabet
+    );
+
+    is(KG_TO_MG, 10**6);
+    is(A, "A");
+    is(B, "B");
+    is(C, "C");
+    like(PI, qr/^3\.14/);
+
+And running it gives:
+
+    $ perl -Ilib t/synopsis_tags.t
+    ok 1
+    ok 2
+    ok 3
+    ok 4
+    ok 5
+    1..5
+
 =head1 DESCRIPTION
 
 This is a utility to write lazy exporters of constant
@@ -493,8 +655,8 @@ subroutines. It's not meant to be a user-facing constant exporting
 API, it's something you use to write user-facing constant exporting
 APIs.
 
-There's dozens of similar constant defining modules and exporters on
-the CPAN, why did I need to write this one?
+There's dozens of modules on the CPAN that define constants in one way
+or another, why did I need to write this one?
 
 =head2 It's lazy
 
@@ -503,11 +665,18 @@ only once for the lifetime of the process (not once per importer or
 whatever), and we only call the callbacks lazily if someone actually
 requests that a constant of ours be defined.
 
-This makes it easy to have one file that runs in different
-environments and generates some subset of its constants with a module
-that you may not want to use, or may not be available in all your
-environments. You can just C<require> it in the callback that
-generates the constant that requires it.
+This makes it easy to have one constant exporting module that runs in
+different environments, and generates some subset of its constants
+depending on what the program that's using it actually needs.
+
+Some data that you may want to turn into constants may require modules
+that aren't available everywhere, queries to databases that aren't
+available everywhere, or make certain assumptions about the
+environment they're running under that may not be true across all your
+environments.
+
+By only defining those constants you actually need via callbacks that
+you provide managing all these special-cases becomes a lot easier.
 
 =head2 It makes it easier to manage creating constants that require other constants
 
@@ -516,17 +685,20 @@ environment, and a bunch of other constants that are defined
 differently if the dev environment constant is true.
 
 Now say you have several hundred constants like that, managing the
-inter-dependencies and that everything is defined in the right order
-quickly gets messy.
+inter-dependencies and ensuring that they're all defined in the right
+order with dependencies before dependents quickly gets messy.
 
 Constant::Import::Lazy takes away all this complexity. When you define
 a constant you get a callback object that can give you the value of
-other constants, and will either generate them if they haven't been
-generated, or look them up in the symbol table if they have.
+other constants.
+
+When you look up another constant we'll either generate it if it
+hasn't been materialized yet, look it up in the symbol table if it
+has.
 
 Thus we end up with a Makefile-like system where you can freely use
-whatever other constants you like when defining your constants, just
-be careful not to introduce circular dependencies.
+whatever other constants you like when defining your constants, you
+only have to be careful not to introduce circular dependencies.
 
 =head1 API
 
@@ -546,9 +718,11 @@ latter, i.e. C<CONST => sub {...}> becomes C<CONST => { call => sub {
 
 The subroutine we'll call with a L<context
 object|Constant::Export::Lazy/"CONTEXT OBJECT"> to fleshen the
-constant. It's guaranteed that this sub will only ever be called once
-for the lifetime of the process, except if you manually call it
-multiple times during an L</override>.
+constant.
+
+It's guaranteed that this sub will only ever be called once for the
+lifetime of the process, except if you manually call it multiple times
+during an L</override>.
 
 =head3 options (local)
 
@@ -564,6 +738,29 @@ L</constants>.
 
 The following options are supported:
 
+=head3 buildargs
+
+A callback that can only be supplied as a global option. If you
+provide this the callback we'll call it to munge any parameters to
+import we might get. This can be used (as shown in the
+L<synopsis|/SYNOPSIS>) to strip or map parameters to e.g. implement
+support for C<%EXPORT_TAGS>, or to do any other arbitrary mapping.
+
+This callback will be called with a reference to the parameters passed
+to import, and for convenience with the L<constants|/constants> with
+the hash you provided (e.g. for introspecting the stashes of
+constants, see the L<synopsis|/SYNOPSIS> example.
+
+This is expected to return an array with a list of constants to
+import, or the empty list if we should discard the return value of
+this callback and act is if it wasn't present at all.
+
+This plays nice with the L</wrap_existing_import> parameter. When it's
+in force any constant names (or tag names, or whatever) you return
+that we don't know about ourselves we'll pass to the fallback import
+subroutine we're wrapping as we would if buildargs hadn't been
+defined.
+
 =head3 wrap_existing_import
 
 A boolean that can only be supplied as a global option. If you provide
@@ -580,6 +777,17 @@ combination of L<Exporter> to export a bunch of L<constant> constants
 without having to port them all over to C<Constant::Export::Lazy> at
 the same time. This allows you to do so incrementally.
 
+For convenience we also support calling these foreign subroutines with
+C<<$ctx->call($name)>>. This is handy because when migrating an
+existing package you can already start calling existing constants with
+our interface, and then when you migrate those constants over you
+won't have to change any of the old code.
+
+We'll handle calling subroutines generated with perl's own
+L<constant.pm|constant> (including "list" constants), but we'll die in
+C<call> if we call a foreign subroutine that returns more than one
+value.
+
 =head3 override
 
 This callback can be defined either globally or locally and will be
@@ -635,6 +843,52 @@ object|Constant::Export::Lazy/"CONTEXT OBJECT">'s C<stash> method
 C</after> calls relevant to its scope, i.e. global if you define it
 globally, otherwise local if it's defined locally.
 
+=head3 private_name_munger
+
+This callback can be defined either globally or locally. When it's
+provided it'll be used to munge the internal name of the subroutine we
+define in the exporting package.
+
+This allows for preventing the anti-pattern of user code not importing
+constants before using them. To take the example in the
+L<synopsis|/SYNOPSIS> it's for preventing C<My::Constants::PI> and
+C<My::User::Code::PI> interchangeably, using this facility we can
+change C<My::Constants::PI> to
+e.g. C<My::Constants::SOME_OPAQUE_VALUE_PI>.
+
+This is useful because users used to using other constant modules
+might be in the habit of using non-imported and imported names
+interchangeably.
+
+This is fine when the constant exporting module isn't lazy, however
+with Constant::Export::Lazy this relies on someone else having
+previously defined the constant at a distance, and if that someone
+goes away this'll silently turn into an error at a distance.
+
+By using the C<private_name_munger> option you can avoid this
+happening in the first place by specifying a subroutine like:
+
+    private_name_munger => sub {
+        my ($gimme) = @_;
+
+        # We guarantee that these constants are always defined by us,
+        # and we don't want to munge them because legacy code calls
+        # them directly for historical reasons.
+        return if $gimme =~ /^ALWAYS_DEFINED_/;
+
+        state $now = time();
+        return $gimme . '_TIME_' . $now;
+    },
+
+Anyone trying to call that directly from your exporting package as
+opposed to importing into their package will very quickly discover
+that it doesn't work.
+
+Because this is called really early on this routine doesn't get passed
+a C<$ctx> object, just the name of the constant you might want to
+munge. To skip munging it return the empty list, otherwise return a
+munged name to be used in the private symbol table.
+
 =head1 CONTEXT OBJECT
 
 As discussed above we pass around a context object to all callbacks
